/*
Copyright (c) Members of the EGEE Collaboration. 2004. 
See http://www.eu-egee.org/partners/ for details on the copyright
holders.  

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 
 */

package org.glite.security.util;

import java.io.File;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.text.ParseException;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.glite.security.util.namespace.EUGridNamespaceFormat;
import org.glite.security.util.namespace.LegacyNamespaceFormat;
import org.glite.security.util.namespace.NamespaceFormat;

/**
 * A class for representing and handling a trust anchor.
 * 
 * This class will maintain the trust anchor data and can be used to poll for
 * updates of the information. I handles the CRL and namespace data in addition
 * to the CA certificate.
 * 
 * @author joni.hahkala@cern.ch
 * 
 */

public class FullTrustAnchor {
    /** Logging facility. */
    private static final Logger LOGGER = Logger.getLogger(FullTrustAnchor.class);

    public static final String IGTF_NAMESPACE_ENDING = ".namespaces";
    public static final String GLOBUS_NAMESPACE_ENDING = ".signing_policy";
    public static final String CRL_FILE_ENDING_PREFIX = ".r";

    static FileCertReader s_certReader;

    static {
        try {
            s_certReader = new FileCertReader();
        } catch (CertificateException e) {
            throw new RuntimeException("Security provider initialization failed: " + e.getMessage(), e);
        }
    }

    /** The 8 byte hash of the DN of this CA. */
    public String m_caHash;

    /** The filename of the CA files without the '.' and ending. */
    public String m_baseFilename;

    /** The running number of the CA, used for the ending. */
    public int m_caNumber;

    /** The CA certificate. */
    public X509Certificate m_caCert;

    /** The time the CA file was last modified. */
    public long m_caModified;

    /** The CRL of the CA. */
    public X509CRL m_crl;

    /** The time the CRL file was last modified. */
    public long m_crlModified;

    /** The namespace object of this CA. */
    public NamespaceFormat m_namespace;

    /** The filename of the namespace in use. */
    public String m_namespaceFilename;

    /** The time the namespace file was last modified. */
    public long m_namespaceModified;

    /** The time the CA dir was last polled for changes. */
    public long m_lastUpdateCheck;

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("TrustAnchor hash: ");
        buf.append(m_caHash);
        buf.append(" DN: ");
        buf.append(DNHandler.getSubject(m_caCert).toString());
        buf.append("\n modified: ");
        buf.append(m_caModified);
        buf.append("\n CRL: ");
        buf.append(m_crl);
        buf.append("\n crl modified: ");
        buf.append(m_crlModified);
        buf.append("\n nameSpace from " + m_namespaceFilename + ": ");
        buf.append(m_namespace);
        buf.append("\n namespace modified: ");
        buf.append(m_namespaceModified);
        buf.append("\n last update check: ");
        buf.append(m_lastUpdateCheck);
        return buf.toString();
    }

    /**
     * Initializes the TrustAnchor from the filename given, loading the
     * corresponding CRL and namespace definition.
     * 
     * @param caFilename
     *            The filename of the CA file.
     * @throws IOException
     *             Thrown if the file opening or reading fails.
     * @throws CertificateException
     *             Thrown in case the Certificate parsing fails.
     * @throws ParseException
     *             Thrown in case the namespace parsing fails.
     */
    FullTrustAnchor(String caFilename) throws IOException, CertificateException, ParseException {
        initCAInfo(caFilename);
    }

    /**
     * Initializes the TrustAnchor from the filename given, loading the
     * corresponding CRL and namespace definition.
     * 
     * @param caFilename
     *            The filename of the CA file.
     * @throws IOException
     *             Thrown if the file opening or reading fails.
     * @throws CertificateException
     *             Thrown in case the Certificate parsing fails.
     * @throws ParseException
     *             Thrown in case the namespace parsing fails.
     */
    void initCAInfo(String caFilename) throws IOException, CertificateException, ParseException {

        if (caFilename == null || caFilename.length() == 0) {
            throw new IOException("Can't initialize a trustanchor without filename.");
        }

        CAFilenameSplitter parts = CAFilenameSplitter.splitCAFilename(caFilename);

        m_caHash = parts.m_hash;
        m_caNumber = parts.m_number;
        m_baseFilename = parts.m_baseFilename;

        loadCACert(m_baseFilename + "." + m_caNumber);

        // allow CRL loading to fail, CA will most likely be disabled though.
        try {
            loadCRL(m_baseFilename + CRL_FILE_ENDING_PREFIX + m_caNumber);
        } catch (Exception e) {
            LOGGER.warn("CRL loading for CA " + m_baseFilename + "." + m_caNumber
                    + " failed, depending on configuration the certificates from the CA might be refused.", e);
        }

        // allow namespace loading to fail, namespaces will not be restricted in
        // that case.
        try {
            loadNamespace(m_baseFilename);
        } catch (Exception e) {
            LOGGER.warn("Namespace restrictions loading for CA " + m_baseFilename + "." + m_caNumber
                    + " failed, namespace is not restricted.", e);
        }
    }

    /**
     * Loads the CA cert and sets the modified time.
     * 
     * @param filename
     *            The name of the file to load.
     * @throws CertificateException
     *             Thrown in case the cert parsing fails.
     * @throws IOException
     *             Thrown in case the file opening or reading fails.
     */
    void loadCACert(String filename) throws CertificateException, IOException {
        Vector certs = s_certReader.readCerts(filename);
        m_caCert = (X509Certificate) certs.get(0); // only support one CA
        // cert on one file.
        File file = new File(filename);
        m_caModified = file.lastModified();

    }

    /**
     * Loads the CRL and sets the modified time.
     * 
     * @param filename
     *            The name of the file to load.
     * @throws CertificateException
     *             Thrown in case the CRL parsing fails.
     * @throws IOException
     *             Thrown in case the file opening or reading fails.
     */
    void loadCRL(String filename) throws CertificateException, IOException {
        Vector certs = s_certReader.readCRLs(filename);
        m_crl = (X509CRL) certs.get(0); // only support one CA cert on one file.
        File file = new File(filename);
        m_crlModified = file.lastModified();
    }

    /**
     * Loads the namespace of this CA, trying first to find the IGTF format and
     * the the legacy format.
     * 
     * @param filename
     *            The base filename of the file to load.
     * @throws CertificateException
     *             Thrown in case the namespace parsing fails.
     * @throws IOException
     *             Thrown in case the file opening or reading fails.
     */
    void loadNamespace(String filenameBase) throws IOException, ParseException {
        // try to load first the new namespace format
        File file = new File(filenameBase + IGTF_NAMESPACE_ENDING);

        if (file.exists()) {
            try {
                m_namespaceFilename = filenameBase + IGTF_NAMESPACE_ENDING;
                m_namespace = new EUGridNamespaceFormat();
                m_namespace.parse(m_namespaceFilename);
                m_namespaceModified = file.lastModified();
                LOGGER.debug("loaded: " + m_namespaceFilename);
            } catch (ParseException e) { // fallback to the old globus format in
                // case loading fails
                LOGGER.warn("Parsing of " + m_namespaceFilename + " failed! Falling back to the "
                        + GLOBUS_NAMESPACE_ENDING + " file");
                m_namespaceFilename = filenameBase + GLOBUS_NAMESPACE_ENDING;
                m_namespace = new LegacyNamespaceFormat();
                m_namespace.parse(m_namespaceFilename);
                file = new File(m_namespaceFilename);
                m_namespaceModified = file.lastModified();
            }
        } else { // fallback to the old globus format in case new format is not
            // found
            m_namespaceFilename = filenameBase + GLOBUS_NAMESPACE_ENDING;
            m_namespace = new LegacyNamespaceFormat();
            m_namespace.parse(m_namespaceFilename);
            file = new File(m_namespaceFilename);
            m_namespaceModified = file.lastModified();
            LOGGER.debug("loaded: " + m_namespaceFilename);
        }
    }

    /**
     * Checks whether there is a need to update and reloads the corresponding
     * file if needed. Will fall back to legacy namespace if the IGTF format one
     * is removed.
     * 
     * @throws CertificateException
     *             Thrown in case the CA cert or CRL parsing fails.
     * @throws IOException
     *             Thrown in case a file opening or reading fails.
     * @throws ParseException
     *             Thrown in case the namespace parsing fails.
     * @throws CertificateNotFoundException
     *             Thrown in case the CA cert is removed.
     */
    void checkUpdate() throws CertificateException, IOException, ParseException, CertificateNotFoundException {
        File caFile = new File(m_baseFilename + "." + m_caNumber);
        if (caFile.exists()) {
            if (caFile.lastModified() != m_caModified) {
                LOGGER.debug("CA file changed, reloading it: " + caFile);
                loadCACert(caFile.getAbsolutePath());
            }
        } else {
            throw new CertificateNotFoundException("The CA file " + caFile.getName() + " can't be found anymore.");
        }

        File crlFile = new File(m_baseFilename + CRL_FILE_ENDING_PREFIX + m_caNumber);
        if (crlFile.lastModified() != m_crlModified) {
            LOGGER.debug("CRL file changed, reloading it: " + crlFile.getName());
            loadCRL(crlFile.getAbsolutePath());
        }

        try {
            // if the current namespacefile is of the old format, check whether
            // there is
            // new format and try to load that one.
            if (!m_namespaceFilename.equals(m_baseFilename + IGTF_NAMESPACE_ENDING)) {
                File namespaceFile = new File(m_baseFilename + IGTF_NAMESPACE_ENDING);
                LOGGER.debug("new format namespace found when old format used, trying to load new format: "
                        + namespaceFile.getName());
                loadNamespace(m_baseFilename);
            } else {
                File namespaceFile = new File(m_namespaceFilename);
                if (namespaceFile.lastModified() != m_namespaceModified) {
                    LOGGER.debug("Namespace file changed, reloading it: " + namespaceFile.getName());
                    loadNamespace(m_baseFilename);
                }
            }
        } catch (IOException e) {
            LOGGER.warn("Namespace loading failed, namespace unrestricted, reason:", e);
        }
    }
}
