/*
Copyright (c) Members of the EGEE Collaboration. 2004. 
See http://www.eu-egee.org/partners/ for details on the copyright
holders.  

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 
 */
package org.glite.security.util;

import org.apache.log4j.Logger;

import org.bouncycastle.asn1.DEROutputStream;
import org.bouncycastle.asn1.pkcs.RSAPrivateKeyStructure;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.jce.provider.JCERSAPrivateCrtKey;

import org.bouncycastle.openssl.PEMReader;
import org.bouncycastle.openssl.PasswordFinder;
import org.bouncycastle.util.encoders.Base64;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;

import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.Security;
import java.security.interfaces.RSAPrivateKey;

/**
 * This class is used to read a private key.
 * 
 * @author Joni Hahkala Created on April 8, 2002, 9:26 PM
 */
public class PrivateKeyReader {
    /** The logging facility. */
    private static final Logger LOGGER = Logger.getLogger(PrivateKeyReader.class);

    /** The carriage return char. */
    private static final byte CARR = '\r';

    /** The newline char. */
    private static final byte NL = '\n';

    /** The length of internal buffer. */
    private static final int BUF_LEN = 1000;

    /** The length of marking. */
    private static final int MARK_LEN = 10000;

    private static final int LINE_LEN = 64;

    static {
        if (Security.getProvider("BC") == null) {
            Security.addProvider(new BouncyCastleProvider());
        }
    }

    /**
     * Reads the private key form the stream that was given in the constructor.
     * After the reading, the file mark is at the end of the key.
     * 
     * @param bin
     *            The BufferedInputStream to read the key from.
     * @param finder
     *            The password finder class to use for requesting the password.
     * @return Returns the private key read.
     * @throws IOException
     *             Thrown if the key reading fails.
     */
    public static PrivateKey read(final BufferedInputStream bin, final PasswordFinder finder) throws IOException {
        try {
            byte[] b = new byte[BUF_LEN];
            bin.mark(MARK_LEN);

            PEMReader pemReader = new PEMReader(new InputStreamReader(bin), finder, "BC");
            Object o = pemReader.readObject();

//            LOGGER.debug("read the keypair" + o);

            KeyPair pair = (KeyPair) o;
            PrivateKey privateKey = pair.getPrivate();

//            LOGGER.debug("the private key is " + privateKey);

            bin.reset();

            String line;
            int num;

            do {
                num = bin.read(b);
                line = new String(b);
                bin.reset();
                skipLine(bin);
                bin.mark(BUF_LEN);
            } while ((num > 0) && !line.startsWith("-----END"));

            bin.mark(MARK_LEN);

            return privateKey;
        } catch (IOException e) {
            LOGGER.error("Error while reading private key from a file. Exception: " + e.getClass().getName()
                    + " message:" + e.getMessage());

            // e.printStackTrace();
            throw e;

            // throw new KeyException("Error while reading private key from
            // file: " + e.getClass().getName()+ " message:" + e.getMessage());
        }
    }

    /**
     * Reads the private key form the stream that was given in the constructor.
     * After the reading, the file mark is at the end of the key.
     * 
     * @param bin
     *            The BufferedInputStream to read the key from.
     * @return Returns the private key read.
     * @throws IOException
     *             Thrown if the key reading fails.
     */
    public static PrivateKey read(final BufferedInputStream bin) throws IOException {
        return read(bin, (PasswordFinder) null);
    }

    /**
     * Reads the private key form the stream that was given in the constructor.
     * After the reading, the file mark is at the end of the key.
     * 
     * @param bin
     *            The BufferedInputStream to read the key from.
     * @param passwd
     *            The password to use to access the private key.
     * @return Returns the private key read.
     * @throws IOException
     *             Thrown if the key reading fails.
     */
    public static PrivateKey read(final BufferedInputStream bin, final String passwd) throws IOException {
        if (passwd == null) {
            return read(bin, (PasswordFinder) null);
        } else {
            return read(bin, new Password(passwd.toCharArray()));
        }
    }

    /**
     * Skips to the next line.
     * 
     * @param stream
     *            the input stream.
     * 
     * @throws IOException
     *             thrown in case of read error.
     */
    public static void skipLine(final BufferedInputStream stream) throws IOException {
        byte[] b = new byte[BUF_LEN]; // the byte buffer
        stream.mark(BUF_LEN + 2); // mark the beginning

        int num = 0;

        while (stream.available() > 0) { // check that there are still
            // something to read
            num = stream.read(b); // read bytes from the file to the byte
            // buffer

            int i = 0;

            while ((i < num) && (b[i] != CARR) && (b[i] != NL)) { // skip
                // untill
                // the first
                // newline
                // or
                // carriage
                // break
                i++;
            }

            stream.reset(); // rewind the file to the beginning of the last read

            if ((b[i] == CARR) || (b[i] == NL)) {
                // String test = new String(b, 0, i);
                // System.out.println("Skipping: " + test);
                stream.skip(i);
                stream.mark(BUF_LEN + 2); // mark the new position

                break;
            }

            stream.skip(BUF_LEN);
            stream.mark(BUF_LEN + 2); // mark the new position
        }

        num = stream.read(b);

        // LOGGER.debug("Remains: " + new String(b, 0, num));
        if ((b[0] != CARR) && (b[0] != NL)) {
            LOGGER.error("No newline char found when trying to skip line");
            throw new IOException("No newline char found when trying to skip line");
        }

        stream.reset();

        // check for carriage return-newline combination
        if ((b[1] == CARR) || ((b[1] == NL) && (b[0] != b[1]))) { 
            stream.skip(2);
        } else {
            stream.skip(1);
        }

        stream.mark(BUF_LEN + 2);

        return;
    }

    /**
     * Return a PKCS1v2 representation of the key. The sequence returned
     * represents a full PrivateKeyInfo object.
     * 
     * @param inKey
     *            the key to encode.
     * @return a PKCS1v2 representation of the key.
     */
    public static byte[] getEncoded(PrivateKey inKey) {
        JCERSAPrivateCrtKey key;

        if (inKey instanceof JCERSAPrivateCrtKey) {
            key = (JCERSAPrivateCrtKey) inKey;
        } else {
            throw new IllegalArgumentException("Argument was:" + inKey + " Expected: JCERSAPrivateCrtKey");
        }

        ByteArrayOutputStream bOut = new ByteArrayOutputStream();
        DEROutputStream dOut = new DEROutputStream(bOut);
        RSAPrivateKeyStructure info = new RSAPrivateKeyStructure(key.getModulus(), key.getPublicExponent(), key
                .getPrivateExponent(), key.getPrimeP(), key.getPrimeQ(), key.getPrimeExponentP(), key
                .getPrimeExponentQ(), key.getCrtCoefficient());

        try {
            dOut.writeObject(info);
            dOut.close();
        } catch (IOException e) {
            throw new RuntimeException("Error encoding RSA public key");
        }

        return bOut.toByteArray();
    }

    /**
     * Return a PEM formatted PKCS1v2 representation of the key. The sequence
     * returned represents a full PrivateKeyInfo object. Had to re-implement as
     * the other PEM encoders produce wrong version of the key.
     * 
     * @param inKey
     *            the key to encode.
     * @return a PEM format PKCS1v2 representation of the key.
     */
    public static String getPEM(final PrivateKey inKey) {
        byte[] bytes = getEncoded(inKey);
        StringBuffer buffer = new StringBuffer();
        if (inKey instanceof RSAPrivateKey) {

            buffer.append("-----BEGIN RSA PRIVATE KEY-----\n");
            String keyPEM = new String(Base64.encode(bytes));

            // split the PEM data into LINE_LEN long lines.
            for (int i = 0; i < keyPEM.length(); i += LINE_LEN) {
                if (keyPEM.length() < i + LINE_LEN) {
                    buffer.append(keyPEM.substring(i, keyPEM.length()));

                } else {
                    buffer.append(keyPEM.substring(i, i + LINE_LEN));
                }
                buffer.append("\n");
            }
            buffer.append("-----END RSA PRIVATE KEY-----\n");
            return buffer.toString();
        }
        throw new IllegalArgumentException(
                "Trying to get PEM format string of non-RSA private key, while only RSA is supported. Class was: "
                        + inKey.getClass().getName());
    }
}
