/*
Copyright (c) Members of the EGEE Collaboration. 2004. 
See http://www.eu-egee.org/partners/ for details on the copyright
holders.  

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 
 */

package org.glite.security.util;

import java.io.File;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.text.ParseException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.Logger;

/** A class that handles the information obtained from the trust directory.
 * 
 * @author hahkala
 *
 */

public class TrustStorage {
    static final Logger LOGGER = Logger.getLogger(TrustStorage.class.getName());

    /** A hashtable of trust anchors.
     * The key is the ASN1 encoded DN of the CA for speed of finding it.
     * The item in the hashtable is a vector of CAs with that DN.
     * The item is a @see Vector as there can be sevaral CA certs with the 
     * same DN, for example in case of CA cert rollover. They each have their 
     * own CRL etc, so they have to be their own @see FullTrustAnchor.  
     */
    Hashtable<String, Vector<FullTrustAnchor>> m_trustStore = new Hashtable<String, Vector<FullTrustAnchor>>(200);

    /** The path to the trust anchors, CRLs and namespace files. */
    String m_storagePath = null;

    /** Whether the trust anchors are preloaded or not. True if the trustanchors are prealoaded. */
    boolean m_preloaded = false;

    /**
     * Generates a new TrustStorage instance.
     * 
     * @param storagePath The path to the trust anchors, CRLs and namespaces.
     */
    public TrustStorage(String storagePath) throws IOException, CertificateException, ParseException{
        this.m_storagePath = storagePath;
        loadAnchors();
    }


    /**
     * Returns the anchors that correspond to the hash given. 
     * 
     * @param hash The hash used to identify the CA.
     * @return The array of trustanchors that are identified by the hash.
     */
    public FullTrustAnchor[] getAnchors(String hash){
        Vector<FullTrustAnchor> anchors = m_trustStore.get(hash);
        if(anchors == null){
            return null;
        }

        return (FullTrustAnchor[])(anchors.toArray(new FullTrustAnchor[]{}));
    }


    /**
     * Loads all the trustanchors.
     * 
     * @throws IOException Thrown in cas there is problems readig the files.
     * @throws CertificateException Thrown in case there is problems parsing the Certificates.
     * @throws ParseException Thrown in case the namespace file parsing fails.
     */
    public void loadAnchors() throws IOException, CertificateException, ParseException{
        TrustDirHandler dirHandler = new TrustDirHandler(m_storagePath);
        dirHandler.init();

        File files[] = dirHandler.getCAs();

        int i = 0;

        while(i < files.length){
            File file = files[i];

            FullTrustAnchor anchor = new FullTrustAnchor(file.getAbsolutePath());

            Vector<FullTrustAnchor> anchors = m_trustStore.get(anchor.m_caHash);

            if(anchors == null){
                anchors = new Vector<FullTrustAnchor>();
            } 
            anchors.add(anchor);
            m_trustStore.put(anchor.m_caHash, anchors);

            i++;
        }
    }

    /**
     * Checks the trust store. It updates the existing CAs, CRLs and namespace definitions in case the files have changed.
     * It also adds any new CAs along with their CRL and namespaces and removes CAs that have been removed from the storage.
     * 
     * @throws IOException in case there is a problem reading files from the file system.
     * @throws CertificateException in case there is problems parsing the CA certificates or CRLs.
     * @throws ParseException in case there is problems parsing the name space files.
     */
    public void checkUpdate() throws IOException, CertificateException, ParseException{
        TrustDirHandler dirHandler = new TrustDirHandler(m_storagePath);
        dirHandler.init();

        File files[] = dirHandler.getCAs();

        String filenames[] = new String[files.length];

        for(int n = 0; n < files.length; n++){
            filenames[n] = files[n].getAbsolutePath();
        }

        int i = 0;

        // update existing CAs and add new ones.
        while(i < filenames.length){
            String filename = filenames[i];

            // split the CA filename to get the CA hash and number.
            CAFilenameSplitter parts = CAFilenameSplitter.splitCAFilename(filename);

            // check if we already have a CA with that hash.
            Vector<FullTrustAnchor> cas = m_trustStore.get(parts.m_hash);

            // if not, create and add it.
            if(cas == null){
                LOGGER.debug("Adding new anchor from file: " + filename);
                FullTrustAnchor anchor = new FullTrustAnchor(filename);
                Vector<FullTrustAnchor> newVector = new Vector<FullTrustAnchor>();
                newVector.add(anchor);
                m_trustStore.put(parts.m_hash, newVector);
                continue;
            }

            boolean found = false;

            // if we do have CA with that hash, check if one of them has the same number.
            Iterator<FullTrustAnchor> iter = cas.iterator();
            while(iter.hasNext()){
                FullTrustAnchor anchor = iter.next();
                if(anchor.m_caNumber == parts.m_number){
                    anchor.checkUpdate();
                    found = true;
                    break;
                }
            }

            if (!found) {
                // CA with a number used in the file couldn't be found, so add it.
                FullTrustAnchor anchor = new FullTrustAnchor(filename);
                cas.add(anchor);
            }

            i++;
        }

        // remove nonexistiong CAs.
        Iterator<Vector<FullTrustAnchor>> iter = m_trustStore.values().iterator();

        // go through the CA vectors
        while(iter.hasNext()){
            Vector<FullTrustAnchor> anchors = iter.next();
            Iterator<FullTrustAnchor> numberIter = anchors.iterator();
            // go thorugh the CAs in the Vector
            while(numberIter.hasNext()){
                FullTrustAnchor anchor = numberIter.next();
                // try to find the file for CA
                int n;
                for(n = 0; n < filenames.length; n++){
                    // test for match between the trustanchor and filename.
                    if(filenames[n].equals(anchor.m_baseFilename + "." + anchor.m_caNumber)){
                        break;
                    }
                }

                // CA file has been removed, remove it from the trust store.
                if(n == filenames.length){
                    numberIter.remove();
                }
            }
            // check if the CA(s) from the vector were removed.
            if(anchors.isEmpty()){
                iter.remove();
            }
        }
    }



    /**
     * Used to get the list of trust anchors in the storage.
     * 
     * @return The array of trust anchors in the storage.
     */
    public FullTrustAnchor[] getAnchors(){
        if(m_trustStore.isEmpty()){
            return null;
        }
        Enumeration<Vector<FullTrustAnchor>> enumer = m_trustStore.elements();

        Vector<FullTrustAnchor> trustAnchors = new Vector<FullTrustAnchor>();

        while(enumer.hasMoreElements()){
            trustAnchors.addAll(enumer.nextElement());
        }

        if(trustAnchors.isEmpty()){
            return null;
        }

        return (FullTrustAnchor[])trustAnchors.toArray(new FullTrustAnchor[]{});

    }

}
