/*
Copyright (c) Members of the EGEE Collaboration. 2004. 
See http://www.eu-egee.org/partners/ for details on the copyright
holders.  

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 
 */

package org.glite.security.util.namespace;

import java.io.IOException;
import java.security.cert.CertPathValidatorException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import org.glite.security.util.DN;

/**
 * Implements DNChecker interface. Checks a DN against a policy/namespace file.
 * 
 * @author alyu
 */
public class DNCheckerImpl implements DNChecker {

//    private static final Logger LOGGER = Logger.getLogger(DNCheckerImpl.class);

    private NamespaceFormat namespaceFormat = null;

    /**
     * Creates a default DN checker using EU grid namespace format.
     */
    public DNCheckerImpl() {
        this(false);
    }

    /**
     * Creates a default DN Checker using legacy policy format.
     * 
     * @param isLegacy
     *            true if old policy format is to be used
     */
    public DNCheckerImpl(boolean isLegacy) {
        if (isLegacy) {
            namespaceFormat = new LegacyNamespaceFormat();
        } else {
            namespaceFormat = new EUGridNamespaceFormat();
        }
    }

    /**
     * Creates a new DN checker with the specified namespace format.
     * 
     * @param namespaceFormat
     *            a namespace format
     */
    public DNCheckerImpl(NamespaceFormat namespaceFormat) {
        this.namespaceFormat = namespaceFormat;
    }

    /**
     * Reads a namespaces file/policy language file.
     * 
     * @param fileName
     *            the namespaces file or policy language file.
     * @throws IOException
     *             if unsuccessful
     * @throws ParseException
     *             if unsuccessful
     */
    public void read(String fileName) throws IOException, ParseException {
        namespaceFormat.parse(fileName);
    }

    /**
     * Checks the specified subjectDN and issuerDN against previous policies.
     * 
     * @param subjectDN
     *            the subject DN.
     * @param issuerDN
     *            the issuer DN.
     * @param prevPolicies
     *            previous policies or null.
     * @return a set of policies for the issuer and subject DN for the next rounds.
     * @throws CertPathValidatorException in case the DNs violate the namespace.
     */
    public List<NamespacePolicy> check(DN subjectDN, DN issuerDN, List<NamespacePolicy> prevPolicies)
            throws CertPathValidatorException {

    	// the list of policies to forward to the next round in the cert path validation.
        List<NamespacePolicy> newPolicyList = null;
        // the new policies for this round.
        List<NamespacePolicy> policyList = namespaceFormat.getPolices();
        if (prevPolicies != null) {
            // add previous policies
            for (NamespacePolicy p : prevPolicies) {
                policyList.add(p);
            }
        }
        boolean isDenied = false; // match for deny policy.
        boolean issuerDNHasWildcard = false; // issuer wilcard 
        boolean matches = false;
        NamespacePolicy failedPolicy = null;

        for (NamespacePolicy policy : policyList) {
            String subjectPolicyDN = policy.getSubjectDN();
            String issuerPolicyDN = policy.getIssuerDN();

            // Check for wild cards
            issuerPolicyDN = cleanupDN(issuerPolicyDN);
            subjectPolicyDN = cleanupDN(subjectPolicyDN);
            // Check for wild cards
            if (issuerPolicyDN.contains(".*")) {
                // new format wild card is ok
                issuerDNHasWildcard = true;
            }

            if (Pattern.matches(issuerPolicyDN, cleanupDN(issuerDN.getX500().toLowerCase()))) {
                if (Pattern.matches(subjectPolicyDN, cleanupDN(subjectDN.getX500().toLowerCase()))) {
                    if (!policy.isSubjectDNPermitted()) {
                    	failedPolicy = policy;
                        break;
                    }
                    matches = true;
                }
                // add to persistent policy list
                if (newPolicyList == null) {
                    newPolicyList = new ArrayList<NamespacePolicy>();
                }
                newPolicyList.add(policy);
            } else { // if (Pattern.matches(issuerPolicyDN
                if (issuerDNHasWildcard) {
                    if (newPolicyList == null) {
                        newPolicyList = new ArrayList<NamespacePolicy>();
                    }
                    newPolicyList.add(policy);
                    issuerDNHasWildcard = false;
                }
            }
        } // for (

        // check for explicit deny
        if (failedPolicy != null) {
            throw new CertPathValidatorException(subjectDN.getX500() + " is denied in the namespace policy line: " + failedPolicy.getPolicyStatement() + " from file: " + failedPolicy.getFilename());
        }
        
        // check if CA or subject DN were out of policies in case there were any.
        if (!matches && policyList != null && !policyList.isEmpty()) {
            throw new CertPathValidatorException("User DN: " + subjectDN.getX500() + " or issuer DN: " + issuerDN.getX500() + " is outside the namespace policy!");       	
        }

        return newPolicyList;
    }

    private String cleanupDN(String dn) {
        if (dn.contains(".*")) {
            // new format wild card is ok
        } else if (dn.contains("*")) {
            // replace '*' with '.*' which then can be used by
            // java.util.regex.Pattern
            dn = dn.replace("*", ".*");
        }
        if (dn.contains("(")) {
            dn = dn.replace("(", "");
        }
        if (dn.contains(")")) {
            dn = dn.replace(")", "");
        }
        return dn;
    }
}
