package org.glite.security.util.proxy;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

import org.bouncycastle.asn1.ASN1EncodableVector;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;

/**
 * An utility class for defining the allowed address space, used both to define
 * the from and to restrictions.
 * 
 * @author joni.hahkala@cern.ch
 * 
 */
public class ProxyRestrictionData {
    private Vector<GeneralName> m_names = new Vector<GeneralName>();

    /**
     * Parses the restriction data from byte array.
     * 
     * @param bytes
     *            The byte array to parse.
     * @throws IOException
     *             In case there is a problem parsing the certificate.
     */
    public ProxyRestrictionData(byte[] bytes) throws IOException {
        GeneralNames names = new GeneralNames((ASN1Sequence) ASN1Sequence.fromByteArray(bytes));
        GeneralName nameArray[] = names.getNames();
        if (nameArray == null) {
            return;
        }
        for (int i = 0; i < nameArray.length; i++) {
            m_names.add(nameArray[i]);
        }
    }

    /**
     * Adds a new IP addressSpace to the data structure.
     * 
     * @param address
     *            The address space to add to the allowed ip address space.
     *            Example of the format: 192.168.0.0/16. Which equals a
     *            192.168.0.0 with a net mask 255.255.0.0. A single IP address
     *            can be defined as xxx.xxx.xxx.xxx/32. <br>
     *            See <a href="http://www.ietf.org/rfc/rfc4632.txt"> RFC
     *            4632.</a> The restriction is of the format used for
     *            NameConstraints, meaning GeneralName with 8 octets for ipv4
     *            and 32 octets for ipv6 addresses.
     */
    public void addIPAddressWithNetmask(String address) {
        m_names.add(new GeneralName(GeneralName.iPAddress, address));
    }

    /**
     * Returns the GeneralNames structure of the restrictions.
     * 
     * @return The GeneralNames class containing the allowed namespaces.
     */
    public GeneralNames getGeneralNames() {
        if (m_names.isEmpty()) {
            return null;
        }

        ASN1EncodableVector sequenceVector = new ASN1EncodableVector();

        Enumeration<GeneralName> namesEnum = m_names.elements();
        while (namesEnum.hasMoreElements()) {
            sequenceVector.add(namesEnum.nextElement().getDERObject());
        }
        return new GeneralNames(new DERSequence(sequenceVector));
    }

    /**
     * Returns a Vector of IP address spaces as defined in rfc 4632.
     * 
     * @return The vector of address spaces defined in this structure.
     */
    public Vector<String> getIPSpaces() {
        Vector<String> ips = new Vector<String>();
        Enumeration<GeneralName> enumGeneralNames = m_names.elements();
        while (enumGeneralNames.hasMoreElements()) {
            GeneralName item = enumGeneralNames.nextElement();
            if (item.getTagNo() == GeneralName.iPAddress) {
                ips.add(item.toString());
            }
        }
        return ips;
    }

}
