package org.glite.security.util.proxy;

import java.io.IOException;

import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;

/**
 * A class for generating and parsing the proxy tracing extensions. <br>
 * 
 * See <a
 * href="http://www.eugridpma.org/documentation/OIDProxyDelegationTracing.pdf">
 * OIDProxyDelegationTracing documentation. </a> <br>
 * 
 * The proxy tracing extension format is below. It is used to trace the proxy
 * delegation chain by putting in each proxy the url of the service accepting
 * the delegation and the url of the client initiating it. Often the delegation
 * is from service to service, in which case it is easy to use the url of the
 * service. If the initiator of the delegation is a user, then the client should
 * put an url containing the client program as the scheme, the host name or IP
 * address and possibly the username as the path. <br>
 * 
 * At the moment only the URI is supported.
 * 
 * <pre>
 * iGTFProxyTracingIssuerName ::= GeneralNames
 *  iGTFProxyTracingSubjectName ::= GeneralNames
 *  
 *  GeneralNames ::= SEQUENCE SIZE (1..MAX) OF GeneralName
 *  
 *  GeneralName ::= CHOICE {
 *           otherName                       [0]     OtherName,
 *           rfc822Name                      [1]     IA5String,
 *           dNSName                         [2]     IA5String,
 *           x400Address                     [3]     ORAddress,
 *           directoryName                   [4]     Name,
 *           ediPartyName                    [5]     EDIPartyName,
 *           uniformResourceIdentifier       [6]     IA5String,
 *           iPAddress                       [7]     OCTET STRING,
 *           registeredID                    [8]     OBJECT IDENTIFIER}
 *  
 *  OtherName ::= SEQUENCE {
 *           type-id    OBJECT IDENTIFIER,
 *           value      [0] EXPLICIT ANY DEFINED BY type-id }
 *  
 *  EDIPartyName ::= SEQUENCE {
 *           nameAssigner            [0]     DirectoryString OPTIONAL,
 *           partyName               [1]     DirectoryString }
 *  
 *  DirectoryString ::= CHOICE {
 *     teletexString           TeletexString (SIZE (1..maxSize),
 *     printableString         PrintableString (SIZE (1..maxSize)),
 *     universalString         UniversalString (SIZE (1..maxSize)),
 *     bmpString               BMPString (SIZE(1..maxSIZE))
 *  }
 * </pre>
 * 
 * @author joni.hahkala@cern.ch
 */
public class ProxyTracingExtension {

    /** The OID to identify issuer proxy tracing extension. */
    public static final String PROXY_TRACING_ISSUER_EXTENSION_OID = "1.2.840.113612.5.5.1.1.1.1";
    /** The OID to identify subject proxy tracing extension. */
    public static final String PROXY_TRACING_SUBJECT_EXTENSION_OID = "1.2.840.113612.5.5.1.1.1.2";

    /** The OID to identify issuer proxy tracing type. */
    public static final int ISSUER_EXTENSION = 1;
    /** The OID to identify issuer proxy tracing type. */
    public static final int SUBJECT_EXTENSION = 2;

    private GeneralNames m_names = null;
    private GeneralName m_name = null;

    /**
     * Generates a new proxy tracing item from the URL.
     * 
     * @param url The URL to identify the issuer or the subject.
     */
    public ProxyTracingExtension(String url) {
        m_name = new GeneralName(GeneralName.uniformResourceIdentifier, url);
        m_names = new GeneralNames(m_name);
    }

    /**
     * Parses the information in the byte array (GeneralNames ASN1 sequence of
     * GeneralName) into a proxy tracing extension object.
     * 
     * @param bytes The bytes of ASN1 encoded proxy tracing extension.
     * @throws IOException In case the byte array does not contain a valid ASN1
     *         encoded proxy tracing extension.
     */
    public ProxyTracingExtension(byte[] bytes) throws IOException {
        m_names = new GeneralNames((ASN1Sequence) ASN1Sequence.fromByteArray(bytes));
        m_name = (GeneralName) m_names.getNames()[1];
    }

    /**
     * Returns the URL inside the proxy tracing data structure.
     * 
     * @return The URL in String format.
     */
    public String getURL() {
        if (m_name.getTagNo() != GeneralName.uniformResourceIdentifier) {
            return null;
        }

        return m_name.toString();
    }
}
