/*
Copyright (c) Members of the EGEE Collaboration. 2004. 
See http://www.eu-egee.org/partners/ for details on the copyright
holders.  

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 
 */
package org.glite.security.util.proxy;

import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.security.SignatureException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.Vector;

import javax.security.auth.x500.X500Principal;

import org.apache.log4j.Logger;
import org.bouncycastle.asn1.ASN1EncodableVector;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DERObject;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.DERPrintableString;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.DERSet;
import org.bouncycastle.asn1.x509.KeyUsage;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.asn1.x509.X509Name;
import org.bouncycastle.x509.X509V3CertificateGenerator;

/**
 * A generator for RFC 3820 type certificates.
 * 
 * @author joni.hahkala@cern.ch
 * 
 */

public class RFC3820ProxyCertificateGenerator extends X509V3CertificateGenerator {
    /** Logging facility. */
    private static final Logger LOGGER = Logger.getLogger(RFC3820ProxyCertificateGenerator.class);

    private BigInteger m_sn = null;
    private X509Name m_basename = null;
    private X509Name m_newDN = null;
    private boolean m_dnSet = false;
    private String m_policyOID = null;
    private DEROctetString m_policyOctets = null;
    private int m_pathLenLimit = ProxyCertInfoExtension.UNLIMITED;

    /**
     * Generates a new proxy DN based on the basename. Takes the base name that
     * should be given before and adds a CN=[serial number] to the end of the
     * basename. If the serial number is not given, a new one is generated.
     */
    // have to use stuff from BC...
    @SuppressWarnings("unchecked")
    private void generateDN() {
        if (m_basename == null) {
            throw new IllegalArgumentException("generateDN: no basename given, can't generate DN.");
        }

        if (m_sn == null) {
            generateCNAndSerial();
        }

        // generate new cn part
        ASN1EncodableVector newCnPart = new ASN1EncodableVector();
        newCnPart.add(X509Name.CN);
        newCnPart.add(new DERPrintableString(m_sn.toString()));

        // copy the RDNs to a new vector so that the new part can be added.
        ASN1Sequence subjectSequence = (ASN1Sequence) m_basename.getDERObject();
        Enumeration<DERObject> subjectParts = (Enumeration<DERObject>) subjectSequence.getObjects();

        ASN1EncodableVector subjectVector = new ASN1EncodableVector();

        while (subjectParts.hasMoreElements()) {
            DERObject part = subjectParts.nextElement();
            subjectVector.add(part);
        }

        subjectVector.add(new DERSet(new DERSequence(newCnPart)));

        // transform the vector into a new X509Name
        DERSequence subjDerSeq = new DERSequence(subjectVector);
        m_newDN = new X509Name(subjDerSeq);

        LOGGER.debug("SubjectDN :" + m_newDN);
    }

    /*
     * Sets the serial number of the certificate and also the serial number to
     * be used to generate the subject DN if not given.
     * 
     * @see
     * org.bouncycastle.x509.X509V3CertificateGenerator#setSerialNumber(java
     * .math.BigInteger)
     */
    public void setSerialNumber(BigInteger sn) {
        m_sn = sn;
        super.setSerialNumber(m_sn);
    }

    /**
     * Generates the required serial number. The serial number is used as the
     * serial number of the certificate and as the new CN part to be added to
     * the baseDN.
     */
    private void generateCNAndSerial() {
        if (m_sn == null) {
            SecureRandom rand = new SecureRandom();
            m_sn = new BigInteger(64, rand);
        }
        setSerialNumber(m_sn);
    }

    /**
     * Sets the proxy certificate info extension.
     * 
     * @param policyOID
     *            optional proxy policy oid, if not given, the
     *            "delegate all policy" oid is used.
     * @param policy
     *            optional proxy policy. If no policy is given the oid is
     *            assumed to suffice.
     * @param pathLenLimit
     *            optional number of subproxies the proxy can have. If null, no
     *            restrictions are set.
     */
    public void setProxyCertInfo(String policyOID, DEROctetString policy, int pathLenLimit) {
        m_policyOID = policyOID;
        m_policyOctets = policy;
        m_pathLenLimit = pathLenLimit;
    }

    /**
     * Set the path length limit, the number of length of proxy cert chain this
     * proxy is allowed so sign.
     * 
     * @param pathLenLimit
     *            the number of proxies that can follow this in the cert chain.
     */
    public void setPathLenLimit(int pathLenLimit) {
        m_pathLenLimit = pathLenLimit;
    }

    /**
     * Sets the policy oid. Useful in cases a policy oid is used without actual
     * policy, like when using: ProxyPolicy.INHERITALL_POLICY_OID,
     * ProxyPolicy.INDEPENDENT_POLICY_OID or ProxyPolicy.LIMITED_PROXY_OID.
     * 
     * @param policyOID
     *            The oid of the policy to be used.
     */
    public void setPolicyOID(String policyOID) {
        m_policyOID = policyOID;
    }

    /*
     * Disable the setting of subject using X500Principal as the encoding of the
     * DN must match the DN of the parent certificate and using the
     * X500Principal the encoding information is lost.
     * 
     * @see
     * org.bouncycastle.x509.X509V3CertificateGenerator#setSubjectDN(javax.security
     * .auth.x500.X500Principal)
     */
    public void setSubjectDN(X500Principal dnPrincipal) {
        throw new IllegalArgumentException();
    }

    /*
     * Sets the subject DN. The DN should start with the subject of the parent
     * certificate as gotten by getSubjectDN as the encoding of it must match
     * the encoding of the subject in the parent certificate and must contain an
     * additional CN=[serial number] RDN field at the end of the DN.
     * 
     * @seeorg.bouncycastle.x509.X509V3CertificateGenerator#setSubjectDN(org.
     * bouncycastle.asn1.x509.X509Name)
     */
    // have to use stuff from BC...
    @SuppressWarnings("unchecked")
    public void setSubjectDN(X509Name dn) {
        this.m_newDN = dn;
        super.setSubjectDN(dn);

        // Dig the serial number from the last CN of the DN.
        Vector oids = m_newDN.getOIDs();
        Vector values = m_newDN.getValues();

        if (!X509Name.CN.equals(oids.lastElement())) {
            throw new IllegalArgumentException("Given DN does not end with CN as required of proxy certificate DNs");
        }

        // do not override explicitly set serial number.
        if (m_sn == null) {
            m_sn = new BigInteger((String) values.lastElement().toString());
            setSerialNumber(m_sn);
        }

        m_dnSet = true;
    }

    /**
     * Sets the base name for the subject DN generation. Must be the DN of the
     * parent.
     * 
     * @param basename
     */
    public void setBaseName(X509Name basename) {
        m_basename = basename;
    }

    /**
     * Add the proxy extension to the certificate.
     */
    private void addProxyExtension() {
        ProxyPolicy proxyPolicy;
        proxyPolicy = new ProxyPolicy(m_policyOID, m_policyOctets);

        ProxyCertInfoExtension extension = new ProxyCertInfoExtension(m_pathLenLimit, proxyPolicy);
        addExtension(ProxyCertInfoExtension.PROXY_CERT_INFO_EXTENSION_OID, true, extension);

    }

    /*
     * Generate a new X509 certificate. If no proxy cert info extension is set,
     * set the default "delegate all" policy without path length restriction. If
     * base name is set @see #setBasename and now subjectDN is set, generates a
     * subjectDN based on the basename and serial number. If no serial number is
     * set one is generated.
     * 
     * @see
     * org.bouncycastle.x509.X509V3CertificateGenerator#generate(java.security
     * .PrivateKey)
     */
    public X509Certificate generate(PrivateKey privateKey) throws InvalidKeyException, SignatureException,
            NoSuchAlgorithmException, CertificateEncodingException {

        // Add mandatory proxy extension.
        addProxyExtension();

        // If the DN is not set and basename is defined, generate dn and set it.
        if (!m_dnSet && m_basename != null) {
            generateDN();
            super.setSubjectDN(m_newDN);
        }

        return super.generate(privateKey);
    }
}
